"""Utilities for reading and writing OME-Zarr metadata."""

from ngio.ome_zarr_meta._meta_handlers import (
    ImageMetaHandler,
    LabelMetaHandler,
    find_image_meta_handler,
    find_label_meta_handler,
    find_plate_meta_handler,
    find_well_meta_handler,
    get_image_meta_handler,
    get_label_meta_handler,
    get_plate_meta_handler,
    get_well_meta_handler,
)
from ngio.ome_zarr_meta.ngio_specs import (
    AxesMapper,
    Dataset,
    ImageInWellPath,
    NgffVersions,
    NgioImageMeta,
    NgioLabelMeta,
    NgioPlateMeta,
    NgioWellMeta,
    PixelSize,
    path_in_well_validation,
)

__all__ = [
    "AxesMapper",
    "Dataset",
    "ImageInWellPath",
    "ImageMetaHandler",
    "ImageMetaHandler",
    "LabelMetaHandler",
    "LabelMetaHandler",
    "NgffVersions",
    "NgffVersions",
    "NgioImageMeta",
    "NgioLabelMeta",
    "NgioPlateMeta",
    "NgioWellMeta",
    "PixelSize",
    "find_image_meta_handler",
    "find_label_meta_handler",
    "find_plate_meta_handler",
    "find_well_meta_handler",
    "get_image_meta_handler",
    "get_label_meta_handler",
    "get_plate_meta_handler",
    "get_well_meta_handler",
    "path_in_well_validation",
]



"""Ngio Tables backend implementations."""

from ngio.tables.backends._abstract_backend import AbstractTableBackend, BackendMeta
from ngio.tables.backends._table_backends import (
    ImplementedTableBackends,
    TableBackendProtocol,
)
from ngio.tables.backends._utils import (
    convert_anndata_to_pandas,
    convert_anndata_to_polars,
    convert_pandas_to_anndata,
    convert_pandas_to_polars,
    convert_polars_to_anndata,
    convert_polars_to_pandas,
    normalize_anndata,
    normalize_pandas_df,
    normalize_polars_lf,
)

__all__ = [
    "AbstractTableBackend",
    "BackendMeta",
    "ImplementedTableBackends",
    "TableBackendProtocol",
    "convert_anndata_to_pandas",
    "convert_anndata_to_polars",
    "convert_pandas_to_anndata",
    "convert_pandas_to_polars",
    "convert_polars_to_anndata",
    "convert_polars_to_pandas",
    "normalize_anndata",
    "normalize_pandas_df",
    "normalize_polars_lf",
]



"""Generic class to handle Image-like data in a OME-NGFF file."""

from collections.abc import Collection, Iterable
from typing import Generic, Literal, TypeVar

import zarr

from ngio.common import (
    ArrayLike,
    Dimensions,
    Roi,
    RoiPixels,
    consolidate_pyramid,
    get_pipe,
    roi_to_slice_kwargs,
    set_pipe,
)
from ngio.ome_zarr_meta import (
    AxesMapper,
    Dataset,
    ImageMetaHandler,
    LabelMetaHandler,
    PixelSize,
)
from ngio.tables import RoiTable
from ngio.utils import NgioFileExistsError, ZarrGroupHandler

_image_handler = TypeVar("_image_handler", ImageMetaHandler, LabelMetaHandler)


class AbstractImage(Generic[_image_handler]):
    """A class to handle a single image (or level) in an OME-Zarr image.

    This class is meant to be subclassed by specific image types.
    """

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: _image_handler,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.

        """
        self._path = path
        self._group_handler = group_handler
        self._meta_handler = meta_handler

        self._dataset = self._meta_handler.meta.get_dataset(path=path)
        self._pixel_size = self._dataset.pixel_size

        try:
            self._zarr_array = self._group_handler.get_array(self._dataset.path)
        except NgioFileExistsError as e:
            raise NgioFileExistsError(f"Could not find the dataset at {path}.") from e

        self._dimensions = Dimensions(
            shape=self._zarr_array.shape, axes_mapper=self._dataset.axes_mapper
        )

        self._axes_mapper = self._dataset.axes_mapper

    def __repr__(self) -> str:
        """Return a string representation of the image."""
        return f"Image(path={self.path}, {self.dimensions})"

    @property
    def meta_handler(self) -> _image_handler:
        """Return the metadata."""
        return self._meta_handler

    @property
    def zarr_array(self) -> zarr.Array:
        """Return the Zarr array."""
        return self._zarr_array

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the image."""
        return self.zarr_array.shape

    @property
    def dtype(self) -> str:
        """Return the dtype of the image."""
        return str(self.zarr_array.dtype)

    @property
    def chunks(self) -> tuple[int, ...]:
        """Return the chunks of the image."""
        return self.zarr_array.chunks

    @property
    def dimensions(self) -> Dimensions:
        """Return the dimensions of the image."""
        return self._dimensions

    @property
    def axes_mapper(self) -> AxesMapper:
        """Return the axes mapper of the image."""
        return self._axes_mapper

    @property
    def is_3d(self) -> bool:
        """Return True if the image is 3D."""
        return self.dimensions.is_3d

    @property
    def is_2d(self) -> bool:
        """Return True if the image is 2D."""
        return self.dimensions.is_2d

    @property
    def is_time_series(self) -> bool:
        """Return True if the image is a time series."""
        return self.dimensions.is_time_series

    @property
    def is_2d_time_series(self) -> bool:
        """Return True if the image is a 2D time series."""
        return self.dimensions.is_2d_time_series

    @property
    def is_3d_time_series(self) -> bool:
        """Return True if the image is a 3D time series."""
        return self.dimensions.is_3d_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return True if the image is multichannel."""
        return self.dimensions.is_multi_channels

    @property
    def space_unit(self) -> str | None:
        """Return the space unit of the image."""
        return self.meta_handler.meta.space_unit

    @property
    def time_unit(self) -> str | None:
        """Return the time unit of the image."""
        return self.meta_handler.meta.time_unit

    @property
    def pixel_size(self) -> PixelSize:
        """Return the pixel size of the image."""
        return self._pixel_size

    @property
    def dataset(self) -> Dataset:
        """Return the dataset of the image."""
        return self._dataset

    @property
    def path(self) -> str:
        """Return the path of the image."""
        return self._dataset.path

    def has_axis(self, axis: str) -> bool:
        """Return True if the image has the given axis."""
        self.axes_mapper.get_index("x")
        return self.dimensions.has_axis(axis)

    def get_array(
        self,
        axes_order: Collection[str] | None = None,
        mode: Literal["numpy", "dask", "delayed"] = "numpy",
        **slice_kwargs: slice | int | Iterable[int],
    ) -> ArrayLike:
        """Get a slice of the image.

        Args:
            axes_order: The order of the axes to return the array.
            mode: The mode to return the array.
            **slice_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        return get_pipe(
            array=self.zarr_array,
            dimensions=self.dimensions,
            axes_order=axes_order,
            mode=mode,
            **slice_kwargs,
        )

    def get_roi(
        self,
        roi: Roi,
        axes_order: Collection[str] | None = None,
        mode: Literal["numpy", "dask", "delayed"] = "numpy",
        **slice_kwargs: slice | int | Iterable[int],
    ) -> ArrayLike:
        """Get a slice of the image.

        Args:
            roi: The region of interest to get the array.
            axes_order: The order of the axes to return the array.
            mode: The mode to return the array.
            **slice_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        return get_roi_pipe(
            image=self, roi=roi, axes_order=axes_order, mode=mode, **slice_kwargs
        )

    def set_array(
        self,
        patch: ArrayLike,
        axes_order: Collection[str] | None = None,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> None:
        """Set a slice of the image.

        Args:
            patch: The patch to set.
            axes_order: The order of the axes to set the patch.
            **slice_kwargs: The slices to set the patch.

        """
        set_pipe(
            array=self.zarr_array,
            patch=patch,
            dimensions=self.dimensions,
            axes_order=axes_order,
            **slice_kwargs,
        )

    def set_roi(
        self,
        roi: Roi,
        patch: ArrayLike,
        axes_order: Collection[str] | None = None,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> None:
        """Set a slice of the image.

        Args:
            roi: The region of interest to set the patch.
            patch: The patch to set.
            axes_order: The order of the axes to set the patch.
            **slice_kwargs: The slices to set the patch.

        """
        return set_roi_pipe(
            image=self, roi=roi, patch=patch, axes_order=axes_order, **slice_kwargs
        )

    def build_image_roi_table(self, name: str = "image") -> RoiTable:
        """Build the ROI table for an image."""
        return build_image_roi_table(image=self, name=name)


def consolidate_image(
    image: AbstractImage,
    order: Literal[0, 1, 2] = 1,
    mode: Literal["dask", "numpy", "coarsen"] = "dask",
) -> None:
    """Consolidate the image on disk."""
    target_paths = image._meta_handler.meta.paths
    targets = [
        image._group_handler.get_array(path)
        for path in target_paths
        if path != image.path
    ]
    consolidate_pyramid(
        source=image.zarr_array, targets=targets, order=order, mode=mode
    )


def get_roi_pipe(
    image: AbstractImage,
    roi: Roi,
    axes_order: Collection[str] | None = None,
    mode: Literal["numpy", "dask", "delayed"] = "numpy",
    **slice_kwargs: slice | int | Iterable[int],
) -> ArrayLike:
    """Get a slice of the image.

    Args:
        image: The image to get the ROI.
        roi: The region of interest to get the array.
        axes_order: The order of the axes to return the array.
        mode: The mode to return the array.
        **slice_kwargs: The slices to get the array.

    Returns:
        The array of the region of interest.
    """
    slice_kwargs = roi_to_slice_kwargs(
        roi=roi,
        pixel_size=image.pixel_size,
        dimensions=image.dimensions,
        **slice_kwargs,
    )
    return get_pipe(
        array=image.zarr_array,
        dimensions=image.dimensions,
        axes_order=axes_order,
        mode=mode,
        **slice_kwargs,
    )


def set_roi_pipe(
    image: AbstractImage,
    roi: Roi,
    patch: ArrayLike,
    axes_order: Collection[str] | None = None,
    **slice_kwargs: slice | int | Iterable[int],
) -> None:
    """Set a slice of the image.

    Args:
        image: The image to set the ROI.
        roi: The region of interest to set the patch.
        patch: The patch to set.
        axes_order: The order of the axes to set the patch.
        **slice_kwargs: The slices to set the patch.

    """
    slice_kwargs = roi_to_slice_kwargs(
        roi=roi,
        pixel_size=image.pixel_size,
        dimensions=image.dimensions,
        **slice_kwargs,
    )
    set_pipe(
        array=image.zarr_array,
        patch=patch,
        dimensions=image.dimensions,
        axes_order=axes_order,
        **slice_kwargs,
    )


def build_image_roi_table(image: AbstractImage, name: str = "image") -> RoiTable:
    """Build the ROI table for an image."""
    dim_z, dim_y, dim_x = (
        image.dimensions.get("z", strict=False),
        image.dimensions.get("y"),
        image.dimensions.get("x"),
    )
    image_roi = RoiPixels(
        name=name,
        x=0,
        y=0,
        z=0,
        x_length=dim_x,
        y_length=dim_y,
        z_length=dim_z,
    )
    return RoiTable(rois=[image_roi.to_roi(pixel_size=image.pixel_size)])



from pathlib import Path

import pytest

from ngio import Image, open_image


@pytest.mark.parametrize(
    "zarr_name",
    [
        "test_image_yx.zarr",
        "test_image_cyx.zarr",
        "test_image_zyx.zarr",
        "test_image_czyx.zarr",
        "test_image_c1yx.zarr",
        "test_image_tyx.zarr",
        "test_image_tcyx.zarr",
        "test_image_tzyx.zarr",
        "test_image_tczyx.zarr",
    ],
)
def test_open_image(images_v04: dict[str, Path], zarr_name: str):
    path = images_v04[zarr_name]
    image = open_image(path)
    assert isinstance(image, Image)



"""Utility to read/write OME-Zarr metadata v0.4."""

from ngio.ome_zarr_meta.v04._v04_spec_utils import (
    ngio_to_v04_image_meta,
    ngio_to_v04_label_meta,
    ngio_to_v04_plate_meta,
    ngio_to_v04_well_meta,
    v04_to_ngio_image_meta,
    v04_to_ngio_label_meta,
    v04_to_ngio_plate_meta,
    v04_to_ngio_well_meta,
)

__all__ = [
    "ngio_to_v04_image_meta",
    "ngio_to_v04_label_meta",
    "ngio_to_v04_plate_meta",
    "ngio_to_v04_well_meta",
    "v04_to_ngio_image_meta",
    "v04_to_ngio_label_meta",
    "v04_to_ngio_plate_meta",
    "v04_to_ngio_well_meta",
]



from pathlib import Path

import pytest
import zarr

from ngio.common._pyramid import on_disk_zoom


@pytest.mark.parametrize(
    "order, mode",
    [
        (0, "dask"),
        (1, "dask"),
        (0, "numpy"),
        (1, "numpy"),
        (0, "coarsen"),
        (1, "coarsen"),
    ],
)
def test_on_disk_zooms(tmp_path: Path, order: int, mode: str):
    source = tmp_path / "source.zarr"
    source_array = zarr.open_array(source, shape=(16, 128, 128), dtype="uint8")

    target = tmp_path / "target.zarr"
    target_array = zarr.open_array(target, shape=(16, 64, 64), dtype="uint8")

    on_disk_zoom(source_array, target_array, order=order, mode=mode)



"""Module for handling the /tables group in an OME-NGFF file."""

from typing import Literal, Protocol

from ngio.tables.v1 import FeatureTableV1, MaskingRoiTableV1, RoiTableV1
from ngio.tables.v1._generic_table import GenericTable
from ngio.tables.v1._roi_table import _GenericRoiTableV1
from ngio.utils import (
    AccessModeLiteral,
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)

GenericRoiTable = _GenericRoiTableV1
RoiTable = RoiTableV1
MaskingRoiTable = MaskingRoiTableV1
FeatureTable = FeatureTableV1


class Table(Protocol):
    """Placeholder class for a table."""

    @staticmethod
    def type() -> str | None:
        """Return the type of the table."""
        ...

    @staticmethod
    def version() -> str | None:
        """Return the version of the table."""
        ...

    @property
    def backend_name(self) -> str | None:
        """The name of the backend."""
        ...

    @classmethod
    def _from_handler(
        cls, handler: ZarrGroupHandler, backend_name: str | None = None
    ) -> "Table":
        """Create a new table from a Zarr group handler."""
        ...

    def _set_backend(
        self,
        handler: ZarrGroupHandler,
        backend_name: str | None = None,
    ) -> None:
        """Set the backend store and path for the table."""
        ...

    def consolidate(self) -> None:
        """Consolidate the table on disk."""
        ...


TypedTable = Literal[
    "roi_table", "masking_roi_table", "feature_table", "generic_roi_table"
]


def _unique_table_name(type_name, version) -> str:
    """Return the unique name for a table."""
    return f"{type_name}_v{version}"


class ImplementedTables:
    """A singleton class to manage the available table handler plugins."""

    _instance = None
    _implemented_tables: dict[str, type[Table]]

    def __new__(cls):
        """Create a new instance of the class if it does not exist."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._implemented_tables = {}
        return cls._instance

    def available_implementations(self) -> list[str]:
        """Get the available table handler versions."""
        return list(self._implemented_tables.keys())

    def get_table(
        self,
        type: str,
        version: str,
        handler: ZarrGroupHandler,
        backend_name: str | None = None,
        strict: bool = True,
    ) -> Table:
        """Try to get a handler for the given store based on the metadata version."""
        _errors = {}
        for name, table_cls in self._implemented_tables.items():
            if name != _unique_table_name(type, version):
                continue
            try:
                table = table_cls._from_handler(
                    handler=handler, backend_name=backend_name
                )
                return table
            except Exception as e:
                if strict:
                    raise NgioValidationError(
                        f"Could not load table {name} from handler. Error: {e}"
                    ) from e
                else:
                    _errors[name] = e
        # If no table was found, we can try to load the table from a generic table
        try:
            table = GenericTable._from_handler(
                handler=handler, backend_name=backend_name
            )
            return table
        except Exception as e:
            _errors["generic"] = e

        if len(_errors) == 0:
            raise NgioValidationError(
                f"Could not find a table implementation for {type} v{version}. "
                f"Available tables: {self.available_implementations()}"
            )

        raise NgioValidationError(
            f"Could not load table from any known version. Errors: {_errors}"
        )

    def add_implementation(self, handler: type[Table], overwrite: bool = False):
        """Register a new table handler."""
        table_type = handler.type()
        version = handler.version()
        if table_type is None:
            raise NgioValueError("Table handler must have a type.")

        if version is None:
            raise NgioValueError("Table handler must have a version.")

        table_unique_name = _unique_table_name(table_type, version)
        if table_unique_name in self._implemented_tables and not overwrite:
            raise NgioValueError(
                f"Table handler for {table_unique_name} already exists. "
                "Use overwrite=True to replace it."
            )
        self._implemented_tables[table_unique_name] = handler


def _get_table_type(handler: ZarrGroupHandler) -> str:
    """Get the type of the table from the handler."""
    attrs = handler.load_attrs()
    return attrs.get("type", "None")


def _get_table_version(handler: ZarrGroupHandler) -> str:
    """Get the version of the table from the handler."""
    attrs = handler.load_attrs()
    return attrs.get("fractal_table_version", "None")


class TablesContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler

        # Validate the group
        # Either contains a tables attribute or is empty
        attrs = self._group_handler.load_attrs()
        if len(attrs) == 0:
            # It's an empty group
            pass
        elif "tables" in attrs and isinstance(attrs["tables"], list):
            # It's a valid group
            pass
        else:
            raise NgioValidationError(
                f"Invalid /tables group. "
                f"Expected a single tables attribute with a list of table names. "
                f"Found: {attrs}"
            )

    def _get_tables_list(self) -> list[str]:
        """Create the /tables group if it doesn't exist."""
        attrs = self._group_handler.load_attrs()
        return attrs.get("tables", [])

    def _get_table_group_handler(self, name: str) -> ZarrGroupHandler:
        """Get the group handler for a table."""
        handler = self._group_handler.derive_handler(path=name)
        return handler

    def list_roi_tables(self) -> list[str]:
        """List all ROI tables in the group."""
        _tables = []
        for _type in ["roi_table", "masking_roi_table"]:
            _tables.extend(self.list(_type))
        return _tables

    def list(self, filter_types: str | None = None) -> list[str]:
        """List all labels in the group."""
        tables = self._get_tables_list()
        if filter_types is None:
            return tables

        filtered_tables = []
        for table_name in tables:
            tb_handler = self._get_table_group_handler(table_name)
            table_type = _get_table_type(tb_handler)
            if table_type == filter_types:
                filtered_tables.append(table_name)
        return filtered_tables

    def get(
        self, name: str, backend_name: str | None = None, strict: bool = True
    ) -> Table:
        """Get a label from the group."""
        if name not in self.list():
            raise KeyError(f"Table '{name}' not found in the group.")

        table_handler = self._get_table_group_handler(name)
        table_type = _get_table_type(table_handler)
        table_version = _get_table_version(table_handler)
        return ImplementedTables().get_table(
            type=table_type,
            version=table_version,
            handler=table_handler,
            backend_name=backend_name,
            strict=strict,
        )

    def add(
        self,
        name: str,
        table: Table,
        backend: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """Add a table to the group."""
        existing_tables = self._get_tables_list()
        if name in existing_tables and not overwrite:
            raise NgioValueError(
                f"Table '{name}' already exists in the group. "
                "Use overwrite=True to replace it."
            )

        table_handler = self._group_handler.derive_handler(
            path=name, overwrite=overwrite
        )

        if backend is None:
            backend = table.backend_name

        table._set_backend(
            handler=table_handler,
            backend_name=backend,
        )
        table.consolidate()
        if name not in existing_tables:
            existing_tables.append(name)
            self._group_handler.write_attrs({"tables": existing_tables})


ImplementedTables().add_implementation(RoiTableV1)
ImplementedTables().add_implementation(MaskingRoiTableV1)
ImplementedTables().add_implementation(FeatureTableV1)

###################################################################################
#
# Utility functions to open and write tables
#
###################################################################################


def open_tables_container(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "a",
    parallel_safe: bool = False,
) -> TablesContainer:
    """Open a table handler from a Zarr store."""
    handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    return TablesContainer(handler)


def open_table(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "a",
    parallel_safe: bool = False,
) -> Table:
    """Open a table from a Zarr store."""
    handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    return ImplementedTables().get_table(
        _get_table_type(handler), _get_table_version(handler), handler
    )


def write_table(
    store: StoreOrGroup,
    table: Table,
    backend: str | None = None,
    cache: bool = False,
    mode: AccessModeLiteral = "a",
    parallel_safe: bool = False,
) -> None:
    """Write a table to a Zarr store."""
    handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    table._set_backend(
        handler=handler,
        backend_name=backend,
    )
    table.consolidate()



"""Implementation of the FeatureTableV1 class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

from typing import Literal

import pandas as pd
from pydantic import BaseModel

from ngio.tables.backends import BackendMeta, ImplementedTableBackends
from ngio.tables.backends._utils import normalize_pandas_df
from ngio.utils import NgioValueError, ZarrGroupHandler


class RegionMeta(BaseModel):
    """Metadata for the region."""

    path: str


class FeatureTableMeta(BackendMeta):
    """Metadata for the ROI table."""

    fractal_table_version: Literal["1"] = "1"
    type: Literal["feature_table"] = "feature_table"
    region: RegionMeta | None = None
    instance_key: str = "label"


class FeatureTableV1:
    """Class to represent a feature table.

    This can be used to load any table that does not have
    a specific definition.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame | None = None,
        reference_label: str | None = None,
    ) -> None:
        """Initialize the GenericTable."""
        if reference_label is None:
            self._meta = FeatureTableMeta()
        else:
            path = f"../labels/{reference_label}"
            self._meta = FeatureTableMeta(region=RegionMeta(path=path))

        self._instance_key = "label"
        if dataframe is None:
            self._dataframe = None
        else:
            self._dataframe = normalize_pandas_df(
                dataframe,
                index_key=self._instance_key,
                index_type="int",
                reset_index=False,
            )
        self._table_backend = None

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        num_rows = len(self.dataframe) if self.dataframe is not None else 0
        num_columns = len(self.dataframe.columns) if self.dataframe is not None else 0
        properties = f"num_rows={num_rows}, num_columns={num_columns}"
        if self.reference_label is not None:
            properties += f", reference_label={self.reference_label}"
        return f"FeatureTableV1({properties})"

    @staticmethod
    def type() -> str:
        """Return the type of the table."""
        return "feature_table"

    @staticmethod
    def version() -> str:
        """The generic table does not have a version.

        Since does not follow a specific schema.
        """
        return "1"

    @property
    def backend_name(self) -> str | None:
        """Return the name of the backend."""
        if self._table_backend is None:
            return None
        return self._table_backend.backend_name()

    @property
    def reference_label(self) -> str | None:
        """Return the reference label."""
        path = self._meta.region
        if path is None:
            return None

        path = path.path
        path = path.split("/")[-1]
        return path

    @property
    def dataframe(self) -> pd.DataFrame:
        """Return the table as a DataFrame."""
        if self._dataframe is None and self._table_backend is None:
            raise NgioValueError(
                "The table does not have a DataFrame in memory nor a backend."
            )

        if self._dataframe is None and self._table_backend is not None:
            self._dataframe = self._table_backend.load_as_pandas_df()

        if self._dataframe is None:
            raise NgioValueError(
                "The table does not have a DataFrame in memory nor a backend."
            )
        return self._dataframe

    @dataframe.setter
    def dataframe(self, dataframe: pd.DataFrame) -> None:
        """Set the table as a DataFrame."""
        self._dataframe = normalize_pandas_df(
            dataframe,
            index_key=self._instance_key,
            index_type="int",
            reset_index=False,
        )

    @classmethod
    def _from_handler(
        cls, handler: ZarrGroupHandler, backend_name: str | None = None
    ) -> "FeatureTableV1":
        """Create a new ROI table from a Zarr group handler."""
        meta = FeatureTableMeta(**handler.load_attrs())
        instance_key = "label" if meta.instance_key is None else meta.instance_key
        if backend_name is None:
            backend = ImplementedTableBackends().get_backend(
                backend_name=meta.backend,
                group_handler=handler,
                index_key=instance_key,
                index_type="int",
            )
        else:
            backend = ImplementedTableBackends().get_backend(
                backend_name=backend_name,
                group_handler=handler,
                index_key=instance_key,
                index_type="int",
            )
            meta.backend = backend_name

        if not backend.implements_pandas:
            raise NgioValueError(
                "The backend does not implement the dataframe protocol."
            )

        table = cls()
        table._meta = meta
        table._table_backend = backend
        return table

    def _set_backend(
        self,
        handler: ZarrGroupHandler,
        backend_name: str | None = None,
    ) -> None:
        """Set the backend of the table."""
        instance_key = "label" if self._instance_key is None else self._instance_key
        backend = ImplementedTableBackends().get_backend(
            backend_name=backend_name,
            group_handler=handler,
            index_key=instance_key,
            index_type="int",
        )
        self._meta.backend = backend_name
        self._table_backend = backend

    def consolidate(self) -> None:
        """Write the current state of the table to the Zarr file."""
        if self._table_backend is None:
            raise NgioValueError(
                "No backend set for the table. "
                "Please add the table to a OME-Zarr Image before calling consolidate."
            )

        self._table_backend.write(
            self.dataframe,
            metadata=self._meta.model_dump(exclude_none=True),
            mode="pandas",
        )



"""Tables implementations for fractal_tables v1."""

from ngio.tables.v1._feature_table import FeatureTableV1
from ngio.tables.v1._generic_table import GenericTable
from ngio.tables.v1._roi_table import MaskingRoiTableV1, RoiTableV1

__all__ = ["FeatureTableV1", "GenericTable", "MaskingRoiTableV1", "RoiTableV1"]



from collections.abc import Iterable

import dask.array as da
import numpy as np
import zarr

from ngio.common._dimensions import Dimensions
from ngio.ome_zarr_meta.ngio_specs import AxesTransformation
from ngio.utils import NgioValueError


def _validate_int(value: int, shape: int) -> int:
    if not isinstance(value, int):
        raise NgioValueError(f"Invalid value {value} of type {type(value)}")
    if value < 0 or value >= shape:
        raise NgioValueError(
            f"Invalid value {value}. Index out of bounds for axis of shape {shape}"
        )
    return value


def _validate_iter_of_ints(value: Iterable[int], shape: int) -> list[int]:
    if not isinstance(value, list):
        raise NgioValueError(f"Invalid value {value} of type {type(value)}")
    value = [_validate_int(v, shape=shape) for v in value]
    return value


def _validate_slice(value: slice, shape: int) -> slice:
    start = value.start if value.start is not None else 0
    start = max(start, 0)
    stop = value.stop if value.stop is not None else shape
    return slice(start, stop)


class SliceTransform(AxesTransformation):
    slices: tuple[slice | tuple[int, ...], ...]


def compute_and_slices(
    *,
    dimensions: Dimensions,
    **slice_kwargs: slice | int | Iterable[int],
) -> SliceTransform:
    _slices = {}
    axes_names = dimensions._axes_mapper.on_disk_axes_names
    for axis_name, slice_ in slice_kwargs.items():
        axis = dimensions._axes_mapper.get_axis(axis_name)
        if axis is None:
            raise NgioValueError(
                f"Invalid axis {axis_name}. "
                f"Not found on the on-disk axes {axes_names}. "
                "If you want to get/set a singletorn value include "
                "it in the axes_order parameter."
            )

        shape = dimensions.get(axis.on_disk_name)

        if isinstance(slice_, int):
            slice_ = _validate_int(slice_, shape)
            slice_ = slice(slice_, slice_ + 1)

        elif isinstance(slice_, Iterable):
            slice_ = _validate_iter_of_ints(slice_, shape)
            slice_ = tuple(slice_)

        elif isinstance(slice_, slice):
            slice_ = _validate_slice(slice_, shape)

        elif not isinstance(slice_, slice):
            raise NgioValueError(
                f"Invalid slice definition {slice_} of type {type(slice_)}"
            )
        _slices[axis.on_disk_name] = slice_

    slices = tuple(_slices.get(axis, slice(None)) for axis in axes_names)
    return SliceTransform(slices=slices)


def numpy_get_slice(array: zarr.Array, slices: SliceTransform) -> np.ndarray:
    return array[slices.slices]


def dask_get_slice(array: zarr.Array, slices: SliceTransform) -> da.Array:
    da_array = da.from_zarr(array)
    return da_array[slices.slices]


def numpy_set_slice(
    array: zarr.Array, patch: np.ndarray, slices: SliceTransform
) -> None:
    array[slices.slices] = patch


def dask_set_slice(array: zarr.Array, patch: da.Array, slices: SliceTransform) -> None:
    da.to_zarr(arr=patch, url=array, region=slices.slices)



"""Implementation of a generic table class."""

import pandas as pd
from anndata import AnnData

from ngio.tables.backends import (
    BackendMeta,
    ImplementedTableBackends,
    convert_anndata_to_pandas,
    convert_pandas_to_anndata,
)
from ngio.utils import NgioValueError, ZarrGroupHandler


class GenericTableMeta(BackendMeta):
    """Metadata for the ROI table."""

    fractal_table_version: str | None = None
    type: str | None = None


class GenericTable:
    """Class to a non-specific table.

    This can be used to load any table that does not have
    a specific definition.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame | None = None,
        anndata: AnnData | None = None,
    ) -> None:
        """Initialize the GenericTable."""
        self._meta = GenericTableMeta()
        if dataframe is None and anndata is None:
            raise NgioValueError(
                "Either a DataFrame or an AnnData object must be provided."
            )

        if dataframe is not None and anndata is not None:
            raise NgioValueError(
                "Only one of DataFrame or AnnData object can be provided."
            )

        self._dataframe = dataframe
        self._anndata = anndata

        self.anndata_native = True if anndata is not None else False

        self._table_backend = None

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        if self._dataframe is not None:
            num_rows = len(self.dataframe)
            num_columns = len(self.dataframe.columns)
            prop = f"num_rows={num_rows}, num_columns={num_columns}, mode=dataframe"
        else:
            prop = "mode=anndata"
        return f"GenericTable({prop})"

    @staticmethod
    def type() -> str:
        """Return the type of the table."""
        return "generic"

    @staticmethod
    def version() -> str:
        """The generic table does not have a version.

        Since does not follow a specific schema.
        """
        return "1"

    @property
    def backend_name(self) -> str | None:
        """Return the name of the backend."""
        if self._table_backend is None:
            return None
        return self._table_backend.backend_name()

    @property
    def dataframe(self) -> pd.DataFrame:
        """Return the table as a DataFrame."""
        if self._dataframe is not None:
            return self._dataframe

        if self._anndata is not None:
            return convert_anndata_to_pandas(self._anndata)

        raise NgioValueError("No table loaded.")

    @dataframe.setter
    def dataframe(self, dataframe: pd.DataFrame) -> None:
        """Set the table as a DataFrame."""
        self._dataframe = dataframe
        self.anndata_native = False

    @property
    def anndata(self) -> AnnData:
        """Return the table as an AnnData object."""
        if self._anndata is not None:
            return self._anndata

        if self._dataframe is not None:
            return convert_pandas_to_anndata(
                self._dataframe,
            )
        raise NgioValueError("No table loaded.")

    @anndata.setter
    def anndata(self, anndata: AnnData) -> None:
        """Set the table as an AnnData object."""
        self._anndata = anndata
        self.anndata_native = True

    @classmethod
    def _from_handler(
        cls, handler: ZarrGroupHandler, backend_name: str | None = None
    ) -> "GenericTable":
        """Create a new ROI table from a Zarr group handler."""
        meta = GenericTableMeta(**handler.load_attrs())
        if backend_name is None:
            backend = ImplementedTableBackends().get_backend(
                backend_name=meta.backend,
                group_handler=handler,
                index_key=None,
            )
        else:
            backend = ImplementedTableBackends().get_backend(
                backend_name=backend_name,
                group_handler=handler,
                index_key=None,
            )
            meta.backend = backend_name

        if backend.implements_anndata():
            anndata = backend.load_as_anndata()
            table = cls(anndata=anndata)

        elif backend.implements_pandas():
            dataframe = backend.load_as_pandas_df()
            table = cls(dataframe=dataframe)
        else:
            raise NgioValueError(
                "The backend does not implement the dataframe protocol."
            )

        table._meta = meta
        table._table_backend = backend
        return table

    def _set_backend(
        self,
        handler: ZarrGroupHandler,
        backend_name: str | None = None,
    ) -> None:
        """Set the backend of the table."""
        backend = ImplementedTableBackends().get_backend(
            backend_name=backend_name,
            group_handler=handler,
            index_key=None,
        )
        self._meta.backend = backend_name
        self._table_backend = backend

    def consolidate(self) -> None:
        """Write the current state of the table to the Zarr file."""
        if self._table_backend is None:
            raise NgioValueError(
                "No backend set for the table. "
                "Please add the table to a OME-Zarr Image before calling consolidate."
            )

        if self.anndata_native:
            self._table_backend.write(
                self.anndata,
                metadata=self._meta.model_dump(exclude_none=True),
                mode="anndata",
            )
        else:
            self._table_backend.write(
                self.dataframe,
                metadata=self._meta.model_dump(exclude_none=True),
                mode="pandas",
            )



from collections.abc import Iterable
from typing import Protocol

import pandas as pd

from ngio.utils import NgioTableValidationError


class TableValidator(Protocol):
    def __call__(self, table: pd.DataFrame) -> pd.DataFrame:
        """Validate the table DataFrame.

        A Validator is just a simple callable that takes a
            DataFrame and returns a DataFrame.

        If the DataFrame is valid, the same DataFrame is returned.
        If the DataFrame is invalid, the Validator can either modify the DataFrame
            to make it valid or raise a NgioTableValidationError.

        Args:
            table (pd.DataFrame): The DataFrame to validate.

        Returns:
            pd.DataFrame: The validated DataFrame.

        """
        ...


def validate_table(
    table_df: pd.DataFrame,
    validators: Iterable[TableValidator] | None = None,
) -> pd.DataFrame:
    """Validate the table DataFrame.

    Args:
        table_df (pd.DataFrame): The DataFrame to validate.
        validators (Collection[Validator] | None): A collection of functions
            used to validate the table. Default is None.

    Returns:
        pd.DataFrame: The validated DataFrame.
    """
    validators = validators or []

    # Apply all provided validators
    for validator in validators:
        table_df = validator(table_df)

    return table_df


####################################################################################################
#
# Common table validators
#
####################################################################################################
def validate_columns(
    table_df: pd.DataFrame,
    required_columns: list[str],
    optional_columns: list[str] | None = None,
) -> pd.DataFrame:
    """Validate the columns headers of the table.

    If a required column is missing, a TableValidationError is raised.
    If a list of optional columns is provided, only required and optional columns are
        allowed in the table.

    Args:
        table_df (pd.DataFrame): The DataFrame to validate.
        required_columns (list[str]): A list of required columns.
        optional_columns (list[str] | None): A list of optional columns.
            Default is None.

    Returns:
        pd.DataFrame: The validated DataFrame.
    """
    table_header = table_df.columns
    for column in required_columns:
        if column not in table_header:
            raise NgioTableValidationError(
                f"Could not find required column: {column} in the table"
            )

    if optional_columns is None:
        return table_df

    possible_columns = [*required_columns, *optional_columns]
    for column in table_header:
        if column not in possible_columns:
            raise NgioTableValidationError(
                f"Could not find column: {column} in the list of possible columns. ",
                f"Possible columns are: {possible_columns}",
            )

    return table_df


def validate_unique_index(table_df: pd.DataFrame) -> pd.DataFrame:
    """Validate that the index of the table is unique."""
    if table_df.index.is_unique:
        return table_df

    # Find the duplicates
    duplicates = table_df.index[table_df.index.duplicated()].tolist()
    raise NgioTableValidationError(
        f"Index of the table contains duplicates values. Duplicate: {duplicates}"
    )



from ngio import PixelSize
from ngio.common import Dimensions, Roi
from ngio.ome_zarr_meta.ngio_specs import AxesMapper, Axis


def test_rois():
    roi = Roi(
        name="test",
        x=0.0,
        y=0.0,
        z=0.0,
        x_length=1.0,
        y_length=1.0,
        z_length=1.0,
        unit="micrometer",  # type: ignore
        other="other",  # type: ignore
    )

    assert roi.x == 0.0

    axes = [Axis(on_disk_name="x"), Axis(on_disk_name="y")]
    ax_mapper = AxesMapper(on_disk_axes=axes)
    dims = Dimensions(shape=(30, 30), axes_mapper=ax_mapper)

    pixel_size = PixelSize(x=1.0, y=1.0, z=1.0)
    raster_roi = roi.to_pixel_roi(pixel_size, dims)

    assert raster_roi.to_slices() == {
        "x": slice(0, 1),
        "y": slice(0, 1),
        "z": slice(0, 1),
    }
    assert roi.model_extra is not None
    assert roi.model_extra["other"] == "other"

    world_roi_2 = raster_roi.to_roi(pixel_size)

    assert world_roi_2.x == 0.0
    assert world_roi_2.y == 0.0
    assert world_roi_2.z == 0.0
    assert world_roi_2.x_length == 1.0
    assert world_roi_2.y_length == 1.0
    assert world_roi_2.z_length == 1.0



"""Dimension metadata.

This is not related to the NGFF metadata,
but it is based on the actual metadata of the image data.
"""

from collections.abc import Collection

from ngio.common._axes_transforms import transform_list
from ngio.ome_zarr_meta import AxesMapper
from ngio.utils import NgioValidationError, NgioValueError


class Dimensions:
    """Dimension metadata."""

    def __init__(
        self,
        shape: tuple[int, ...],
        axes_mapper: AxesMapper,
    ) -> None:
        """Create a Dimension object from a Zarr array.

        Args:
            shape: The shape of the Zarr array.
            axes_mapper: The axes mapper object.
        """
        self._shape = shape
        self._axes_mapper = axes_mapper

        if len(self._shape) != len(self._axes_mapper.on_disk_axes):
            raise NgioValidationError(
                "The number of dimensions must match the number of axes. "
                f"Expected Axis {self._axes_mapper.on_disk_axes_names} but got shape "
                f"{self._shape}."
            )

    def __str__(self) -> str:
        """Return the string representation of the object."""
        dims = ", ".join(
            f"{ax.on_disk_name}: {s}"
            for ax, s in zip(self._axes_mapper.on_disk_axes, self._shape, strict=True)
        )
        return f"Dimensions({dims})"

    def get(self, axis_name: str, strict: bool = True) -> int:
        """Return the dimension of the given axis name.

        Args:
            axis_name: The name of the axis (either canonical or non-canonical).
            strict: If True, raise an error if the axis does not exist.
        """
        index = self._axes_mapper.get_index(axis_name)
        if index is None and strict:
            raise NgioValueError(f"Axis {axis_name} does not exist.")
        elif index is None:
            return 1
        return self._shape[index]

    def has_axis(self, axis_name: str) -> bool:
        """Return whether the axis exists."""
        index = self._axes_mapper.get_axis(axis_name)
        if index is None:
            return False
        return True

    def get_shape(self, axes_order: Collection[str]) -> tuple[int, ...]:
        """Return the shape in the given axes order."""
        transforms = self._axes_mapper.to_order(axes_order)
        return tuple(transform_list(list(self._shape), 1, transforms))

    def get_canonical_shape(self) -> tuple[int, ...]:
        """Return the shape in the canonical order."""
        transforms = self._axes_mapper.to_canonical()
        return tuple(transform_list(list(self._shape), 1, transforms))

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        return str(self)

    @property
    def on_disk_shape(self) -> tuple[int, ...]:
        """Return the shape as a tuple."""
        return tuple(self._shape)

    @property
    def is_time_series(self) -> bool:
        """Return whether the data is a time series."""
        if self.get("t", strict=False) == 1:
            return False
        return True

    @property
    def is_2d(self) -> bool:
        """Return whether the data is 2D."""
        if self.get("z", strict=False) != 1:
            return False
        return True

    @property
    def is_2d_time_series(self) -> bool:
        """Return whether the data is a 2D time series."""
        return self.is_2d and self.is_time_series

    @property
    def is_3d(self) -> bool:
        """Return whether the data is 3D."""
        return not self.is_2d

    @property
    def is_3d_time_series(self) -> bool:
        """Return whether the data is a 3D time series."""
        return self.is_3d and self.is_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return whether the data has multiple channels."""
        if self.get("c", strict=False) == 1:
            return False
        return True



import pytest

from ngio.common import Dimensions
from ngio.ome_zarr_meta import AxesMapper
from ngio.ome_zarr_meta.ngio_specs import Axis
from ngio.utils import NgioValidationError, NgioValueError


@pytest.mark.parametrize(
    "axes_names",
    [
        ["x", "y", "z", "c"],
        ["x", "y", "c"],
        ["z", "c", "x", "y"],
        ["t", "z", "c", "x", "y"],
        ["x", "y", "z", "t"],
    ],
)
def test_dimensions(axes_names):
    axes = [Axis(on_disk_name=name) for name in axes_names]
    canonic_dim_dict = dict(zip("tczyx", (2, 3, 4, 5, 6), strict=True))
    dim_dict = {ax: canonic_dim_dict.get(ax, 1) for ax in axes_names}

    shape = tuple(dim_dict.get(ax) for ax in axes_names)
    shape = tuple(s for s in shape if s is not None)

    ax_mapper = AxesMapper(on_disk_axes=axes)
    dims = Dimensions(shape=shape, axes_mapper=ax_mapper)

    assert isinstance(dims.__repr__(), str)

    for ax, s in dim_dict.items():
        assert dims.get(ax) == s

    if dim_dict.get("z", 1) > 1:
        assert dims.is_3d

    if dim_dict.get("c", 1) > 1:
        assert dims.is_multi_channels

    if dim_dict.get("t", 1) > 1:
        assert dims.is_time_series

    if dim_dict.get("z", 1) > 1 and dim_dict.get("t", 1) > 1:
        assert dims.is_3d_time_series

    if dim_dict.get("z", 1) == 1 and dim_dict.get("t", 1) == 1:
        assert dims.is_2d

    if dim_dict.get("z", 1) == 1 and dim_dict.get("t", 1) > 1:
        assert dims.is_2d_time_series

    assert dims.get_canonical_shape() == tuple(dim_dict.get(ax, 1) for ax in "tczyx")

    assert dims.on_disk_shape == shape


def test_dimensions_error():
    axes = [Axis(on_disk_name="x"), Axis(on_disk_name="y")]
    shape = (1, 2, 3)

    with pytest.raises(NgioValidationError):
        Dimensions(shape=shape, axes_mapper=AxesMapper(on_disk_axes=axes))

    shape = (3, 4)
    dims = Dimensions(shape=shape, axes_mapper=AxesMapper(on_disk_axes=axes))

    assert dims.get_shape(axes_order=["c", "x", "y", "z"]) == (1, 3, 4, 1)
    assert dims.get_shape(axes_order=["c", "z", "y", "x"]) == (1, 1, 4, 3)

    with pytest.raises(NgioValueError):
        dims.get("c", strict=True)

    assert not dims.is_3d
    assert not dims.is_multi_channels
    assert not dims.is_time_series
    assert not dims.is_3d_time_series
    assert not dims.is_2d_time_series



from pathlib import Path

import pytest

from ngio.utils import download_ome_zarr_dataset, list_ome_zarr_datasets


def test_list_datasets():
    assert len(list_ome_zarr_datasets()) > 0


def test_fail_download_ome_zarr_dataset(tmp_path: Path):
    tmp_path = Path(tmp_path) / "test_datasets_fail"
    with pytest.raises(ValueError):
        download_ome_zarr_dataset("unknown_dataset", download_dir=tmp_path)

    assert not tmp_path.exists()



"""A module for handling label images in OME-NGFF files."""

from collections.abc import Collection, Iterable
from typing import Literal

from ngio.common import ArrayLike, get_masked_pipe, roi_to_slice_kwargs, set_masked_pipe
from ngio.images.image import Image
from ngio.images.label import Label
from ngio.ome_zarr_meta import ImageMetaHandler, LabelMetaHandler
from ngio.tables import MaskingRoiTable
from ngio.utils import (
    ZarrGroupHandler,
)


class MaskedImage(Image):
    """Placeholder class for a label."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: ImageMetaHandler | None,
        label: Label,
        masking_roi_table: MaskingRoiTable,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.
            label: The label image.
            masking_roi_table: The masking ROI table.

        """
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )
        self._label = label
        self._masking_roi_table = masking_roi_table

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        label_name = self._label.meta.name
        if label_name is None:
            label_name = self._masking_roi_table.reference_label
        return f"MaskedImage(path={self.path}, {self.dimensions}, {label_name})"

    def get_roi(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Collection[str] | None = None,
        mode: Literal["numpy", "dask", "delayed"] = "numpy",
        **slice_kwargs: slice | int | Iterable[int],
    ) -> ArrayLike:
        """Return the array for a given ROI."""
        roi = self._masking_roi_table.get(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi(
            roi=roi, axes_order=axes_order, mode=mode, **slice_kwargs
        )

    def set_roi(
        self,
        label: int,
        patch: ArrayLike,
        zoom_factor: float = 1.0,
        axes_order: Collection[str] | None = None,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> None:
        """Set the array for a given ROI."""
        roi = self._masking_roi_table.get(label)
        roi = roi.zoom(zoom_factor)
        return super().set_roi(
            roi=roi, patch=patch, axes_order=axes_order, **slice_kwargs
        )

    def get_roi_masked(
        self,
        label: int,
        axes_order: Collection[str] | None = None,
        mode: Literal["numpy", "dask", "delayed"] = "numpy",
        zoom_factor: float = 1.0,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> ArrayLike:
        """Return the masked array for a given label."""
        return get_masked_roi_pipe(
            image=self,
            label=label,
            axes_order=axes_order,
            mode=mode,
            zoom_factor=zoom_factor,
            **slice_kwargs,
        )

    def set_roi_masked(
        self,
        label: int,
        patch: ArrayLike,
        axes_order: Collection[str] | None = None,
        zoom_factor: float = 1.0,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> None:
        """Set the masked array for a given label."""
        return set_masked_roi_pipe(
            image=self,
            label=label,
            patch=patch,
            axes_order=axes_order,
            zoom_factor=zoom_factor,
            **slice_kwargs,
        )


class MaskedLabel(Label):
    """Placeholder class for a label."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: LabelMetaHandler | None,
        label: Label,
        masking_roi_table: MaskingRoiTable,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.
            label: The label image.
            masking_roi_table: The masking ROI table.

        """
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )
        self._label = label
        self._masking_roi_table = masking_roi_table

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        label_name = self._label.meta.name
        if label_name is None:
            label_name = self._masking_roi_table.reference_label
        return f"MaskedLabel(path={self.path}, {self.dimensions}, {label_name})"

    def get_roi(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Collection[str] | None = None,
        mode: Literal["numpy", "dask", "delayed"] = "numpy",
        **slice_kwargs: slice | int | Iterable[int],
    ) -> ArrayLike:
        """Return the array for a given ROI."""
        roi = self._masking_roi_table.get(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi(
            roi=roi, axes_order=axes_order, mode=mode, **slice_kwargs
        )

    def set_roi(
        self,
        label: int,
        patch: ArrayLike,
        zoom_factor: float = 1.0,
        axes_order: Collection[str] | None = None,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> None:
        """Set the array for a given ROI."""
        roi = self._masking_roi_table.get(label)
        roi = roi.zoom(zoom_factor)
        return super().set_roi(
            roi=roi, patch=patch, axes_order=axes_order, **slice_kwargs
        )

    def get_roi_masked(
        self,
        label: int,
        axes_order: Collection[str] | None = None,
        mode: Literal["numpy", "dask", "delayed"] = "numpy",
        zoom_factor: float = 1.0,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> ArrayLike:
        """Return the masked array for a given label."""
        return get_masked_roi_pipe(
            image=self,
            label=label,
            axes_order=axes_order,
            mode=mode,
            zoom_factor=zoom_factor,
            **slice_kwargs,
        )

    def set_roi_masked(
        self,
        label: int,
        patch: ArrayLike,
        axes_order: Collection[str] | None = None,
        zoom_factor: float = 1.0,
        **slice_kwargs: slice | int | Iterable[int],
    ) -> None:
        """Set the masked array for a given label."""
        return set_masked_roi_pipe(
            image=self,
            label=label,
            patch=patch,
            axes_order=axes_order,
            zoom_factor=zoom_factor,
            **slice_kwargs,
        )


def get_masked_roi_pipe(
    image: MaskedImage | MaskedLabel,
    label: int,
    axes_order: Collection[str] | None = None,
    mode: Literal["numpy", "dask", "delayed"] = "numpy",
    zoom_factor: float = 1.0,
    **slice_kwargs: slice | int | Iterable[int],
) -> ArrayLike:
    """Return the masked array for a given label."""
    roi = image._masking_roi_table.get(label)
    roi = roi.zoom(zoom_factor)
    slice_kwargs = roi_to_slice_kwargs(
        roi=roi,
        pixel_size=image.pixel_size,
        dimensions=image.dimensions,
        **slice_kwargs,
    )
    return get_masked_pipe(
        array=image.zarr_array,
        label_array=image._label.zarr_array,
        label=label,
        dimensions_array=image.dimensions,
        dimensions_label=image._label.dimensions,
        axes_order=axes_order,
        mode=mode,
        **slice_kwargs,
    )


def set_masked_roi_pipe(
    image: MaskedImage | MaskedLabel,
    label: int,
    patch: ArrayLike,
    axes_order: Collection[str] | None = None,
    zoom_factor: float = 1.0,
    **slice_kwargs: slice | int | Iterable[int],
) -> None:
    """Set the masked array for a given label."""
    roi = image._masking_roi_table.get(label)
    roi = roi.zoom(zoom_factor)
    slice_kwargs = roi_to_slice_kwargs(
        roi=roi,
        pixel_size=image.pixel_size,
        dimensions=image.dimensions,
        **slice_kwargs,
    )
    return set_masked_pipe(
        array=image.zarr_array,
        label_array=image._label.zarr_array,
        label=label,
        patch=patch,
        dimensions_array=image.dimensions,
        dimensions_label=image._label.dimensions,
        axes_order=axes_order,
        **slice_kwargs,
    )



import logging

from ngio.utils._errors import NgioValueError

# Configure the logger
ngio_logger = logging.getLogger("NgioLogger")
ngio_logger.setLevel(logging.ERROR)

# Set up a console handler with a custom format
console_handler = logging.StreamHandler()
formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s - %(name)s - "
    "[%(module)s.%(funcName)s:%(lineno)d]: %(message)s"
)
console_handler.setFormatter(formatter)

# Add the handler to the logger
ngio_logger.addHandler(console_handler)


def set_logger_level(level: str) -> None:
    """Set the logger level.

    Args:
        level: The level to set the logger to.
            Must be one of "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
    """
    if level not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        raise NgioValueError(f"Invalid log level: {level}")

    ngio_logger.setLevel(level)



"""A module for handling the Plate Collection in an OME-Zarr file."""

from typing import Literal, overload

from ngio.images import OmeZarrContainer
from ngio.ome_zarr_meta import (
    ImageInWellPath,
    NgffVersions,
    NgioPlateMeta,
    NgioWellMeta,
    find_plate_meta_handler,
    find_well_meta_handler,
    get_plate_meta_handler,
    get_well_meta_handler,
    path_in_well_validation,
)
from ngio.tables import (
    FeatureTable,
    GenericRoiTable,
    MaskingRoiTable,
    RoiTable,
    Table,
    TablesContainer,
    TypedTable,
)
from ngio.utils import (
    AccessModeLiteral,
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)


def _default_table_container(handler: ZarrGroupHandler) -> TablesContainer | None:
    """Return a default table container."""
    success, table_handler = handler.safe_derive_handler("tables")
    if success and isinstance(table_handler, ZarrGroupHandler):
        return TablesContainer(table_handler)


# Mock lock class that does nothing
class MockLock:
    """A mock lock class that does nothing."""

    def __enter__(self):
        """Enter the lock."""
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the lock."""
        pass


class OmeZarrWell:
    """A class to handle the Well Collection in an OME-Zarr file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler.

        Args:
            group_handler: The Zarr group handler that contains the Well.
        """
        self._group_handler = group_handler
        self._meta_handler = find_well_meta_handler(group_handler)

    def __repr__(self) -> str:
        """Return a string representation of the well."""
        return f"Well(#images: {len(self.paths())})"

    @property
    def meta_handler(self):
        """Return the metadata handler."""
        return self._meta_handler

    @property
    def meta(self):
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def acquisition_ids(self) -> list[int]:
        """Return the acquisitions ids in the well."""
        return self.meta.acquisition_ids

    def paths(self, acquisition: int | None = None) -> list[str]:
        """Return the images paths in the well.

        If acquisition is None, return all images paths in the well.
        Else, return the images paths in the well for the given acquisition.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
        """
        return self.meta.paths(acquisition)

    def get_image_store(self, image_path: str) -> StoreOrGroup:
        """Get the image store from the well.

        Args:
            image_path (str): The path of the image.
        """
        return self._group_handler.get_group(image_path, create_mode=True)

    def get_image_acquisition_id(self, image_path: str) -> int | None:
        """Get the acquisition id of an image in the well.

        Args:
            image_path (str): The path of the image.

        Returns:
            int | None: The acquisition id of the image.
        """
        return self.meta.get_image_acquisition_id(image_path=image_path)

    def get_image(self, image_path: str) -> OmeZarrContainer:
        """Get an image from the well.

        Args:
            image_path (str): The path of the image.

        Returns:
            OmeZarrContainer: The image.
        """
        handler = self._group_handler.derive_handler(image_path)
        return OmeZarrContainer(handler)

    def _add_image(
        self,
        image_path: str,
        acquisition_id: int | None = None,
        strict: bool = True,
        atomic: bool = False,
    ) -> StoreOrGroup:
        """Add an image to an ome-zarr well."""
        image_path = path_in_well_validation(path=image_path)

        if atomic:
            well_lock = self._group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            meta = self.meta.add_image(
                path=image_path, acquisition=acquisition_id, strict=strict
            )
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

        return self._group_handler.get_group(image_path, create_mode=True)

    def atomic_add_image(
        self,
        image_path: str,
        acquisition_id: int | None = None,
        strict: bool = True,
    ) -> StoreOrGroup:
        """Parallel safe version of add_image."""
        return self._add_image(
            image_path=image_path,
            acquisition_id=acquisition_id,
            atomic=True,
            strict=strict,
        )

    def add_image(
        self,
        image_path: str,
        acquisition_id: int | None = None,
        strict: bool = True,
    ) -> StoreOrGroup:
        """Add an image to an ome-zarr well.

        Args:
            image_path (str): The path of the image.
            acquisition_id (int | None): The acquisition id to filter the images.
            strict (bool): Whether to check if the acquisition id is already exists
                in the well. Defaults to True. If False this might lead to
                acquision in a well that does not exist at the plate level.
        """
        return self._add_image(
            image_path=image_path,
            acquisition_id=acquisition_id,
            atomic=False,
            strict=strict,
        )


class OmeZarrPlate:
    """A class to handle the Plate Collection in an OME-Zarr file."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        table_container: TablesContainer | None = None,
    ) -> None:
        """Initialize the LabelGroupHandler.

        Args:
            group_handler: The Zarr group handler that contains the Plate.
            table_container: The tables container that contains plate level tables.
        """
        self._group_handler = group_handler
        self._meta_handler = find_plate_meta_handler(group_handler)
        self._tables_container = table_container

    def __repr__(self) -> str:
        """Return a string representation of the plate."""
        return f"Plate([rows x columns] ({len(self.rows)} x {len(self.columns)}))"

    @property
    def meta_handler(self):
        """Return the metadata handler."""
        return self._meta_handler

    @property
    def meta(self):
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def columns(self) -> list[str]:
        """Return the number of columns in the plate."""
        return self.meta.columns

    @property
    def rows(self) -> list[str]:
        """Return the number of rows in the plate."""
        return self.meta.rows

    @property
    def acquisitions_names(self) -> list[str | None]:
        """Return the acquisitions in the plate."""
        return self.meta.acquisitions_names

    @property
    def acquisition_ids(self) -> list[int]:
        """Return the acquisitions ids in the plate."""
        return self.meta.acquisition_ids

    def _well_path(self, row: str, column: int | str) -> str:
        """Return the well path in the plate."""
        return self.meta.get_well_path(row=row, column=column)

    def _image_path(self, row: str, column: int | str, path: str) -> str:
        """Return the image path in the plate."""
        well = self.get_well(row, column)
        if path not in well.paths():
            raise ValueError(f"Image {path} does not exist in well {row}{column}")
        return f"{self._well_path(row, column)}/{path}"

    def wells_paths(self) -> list[str]:
        """Return the wells paths in the plate."""
        return self.meta.wells_paths

    def images_paths(self, acquisition: int | None = None) -> list[str]:
        """Return the images paths in the plate.

        If acquisition is None, return all images paths in the plate.
        Else, return the images paths in the plate for the given acquisition.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
        """
        images = []
        for well_path, wells in self.get_wells().items():
            for img_path in wells.paths(acquisition):
                images.append(f"{well_path}/{img_path}")
        return images

    def well_images_paths(
        self, row: str, column: int | str, acquisition: int | None = None
    ) -> list[str]:
        """Return the images paths in a well.

        If acquisition is None, return all images paths in the well.
        Else, return the images paths in the well for the given acquisition.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            acquisition (int | None): The acquisition id to filter the images.
        """
        images = []
        well = self.get_well(row=row, column=column)
        for path in well.paths(acquisition):
            images.append(self._image_path(row=row, column=column, path=path))
        return images

    def get_image_acquisition_id(
        self, row: str, column: int | str, image_path: str
    ) -> int | None:
        """Get the acquisition id of an image in a well.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            image_path (str): The path of the image.

        Returns:
            int | None: The acquisition id of the image.
        """
        well = self.get_well(row=row, column=column)
        return well.get_image_acquisition_id(image_path=image_path)

    def get_well(self, row: str, column: int | str) -> OmeZarrWell:
        """Get a well from the plate.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.

        Returns:
            OmeZarrWell: The well.
        """
        well_path = self._well_path(row=row, column=column)
        group_handler = self._group_handler.derive_handler(well_path)
        return OmeZarrWell(group_handler)

    def get_wells(self) -> dict[str, OmeZarrWell]:
        """Get all wells in the plate.

        Returns:
            dict[str, OmeZarrWell]: A dictionary of wells, where the key is the well
                path and the value is the well object.
        """
        wells = {}
        for well_path in self.wells_paths():
            group_handler = self._group_handler.derive_handler(well_path)
            well = OmeZarrWell(group_handler)
            wells[well_path] = well
        return wells

    def get_images(self, acquisition: int | None = None) -> dict[str, OmeZarrContainer]:
        """Get all images in the plate.

        Args:
            acquisition: The acquisition id to filter the images.
        """
        images = {}
        for image_path in self.images_paths(acquisition):
            img_group_handler = self._group_handler.derive_handler(image_path)
            images[image_path] = OmeZarrContainer(img_group_handler)
        return images

    def get_image(
        self, row: str, column: int | str, image_path: str
    ) -> OmeZarrContainer:
        """Get an image from the plate.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            image_path (str): The path of the image.

        Returns:
            OmeZarrContainer: The image.
        """
        image_path = self._image_path(row=row, column=column, path=image_path)
        group_handler = self._group_handler.derive_handler(image_path)
        return OmeZarrContainer(group_handler)

    def get_image_store(
        self, row: str, column: int | str, image_path: str
    ) -> StoreOrGroup:
        """Get the image store from the plate.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            image_path (str): The path of the image.
        """
        well = self.get_well(row=row, column=column)
        return well.get_image_store(image_path=image_path)

    def get_well_images(
        self, row: str, column: str | int, acquisition: int | None = None
    ) -> dict[str, OmeZarrContainer]:
        """Get all images in a well.

        Args:
            row: The row of the well.
            column: The column of the well.
            acquisition: The acquisition id to filter the images.
        """
        images = {}
        for image_paths in self.well_images_paths(
            row=row, column=column, acquisition=acquisition
        ):
            group_handler = self._group_handler.derive_handler(image_paths)
            images[image_paths] = OmeZarrContainer(group_handler)
        return images

    def _add_image(
        self,
        row: str,
        column: int | str,
        image_path: str | None = None,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
        atomic: bool = False,
    ) -> str:
        """Add an image to an ome-zarr plate."""
        if image_path is not None:
            image_path = path_in_well_validation(path=image_path)

        if atomic:
            plate_lock = self._group_handler.lock
        else:
            plate_lock = MockLock()

        with plate_lock:
            meta = self.meta
            meta = meta.add_well(row=row, column=column)
            if acquisition_id is not None:
                meta = meta.add_acquisition(
                    acquisition_id=acquisition_id, acquisition_name=acquisition_name
                )
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

        well_path = self.meta.get_well_path(row=row, column=column)
        group_handler = self._group_handler.derive_handler(well_path)

        if atomic:
            well_lock = group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            attrs = group_handler.load_attrs()
            if len(attrs) == 0:
                # Initialize the well metadata
                # if the group is empty
                well_meta = NgioWellMeta.default_init()
                version = self.meta.plate.version
                version = version if version is not None else "0.4"
                meta_handler = get_well_meta_handler(group_handler, version=version)
            else:
                meta_handler = find_well_meta_handler(group_handler)
                well_meta = meta_handler.meta

            group_handler = self._group_handler.derive_handler(well_path)

            if image_path is not None:
                well_meta = well_meta.add_image(
                    path=image_path, acquisition=acquisition_id, strict=False
                )
            meta_handler.write_meta(well_meta)
            meta_handler._group_handler.clean_cache()

        if image_path is not None:
            return f"{well_path}/{image_path}"
        return well_path

    def atomic_add_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
    ) -> str:
        """Parallel safe version of add_image."""
        if image_path is None:
            raise ValueError(
                "Image path cannot be None for atomic add_image. "
                "If your intent is to add a well, use add_well instead."
            )
        path = self._add_image(
            row=row,
            column=column,
            image_path=image_path,
            acquisition_id=acquisition_id,
            acquisition_name=acquisition_name,
            atomic=True,
        )
        return path

    def add_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
    ) -> str:
        """Add an image to an ome-zarr plate."""
        if image_path is None:
            raise ValueError(
                "Image path cannot be None for atomic add_image. "
                "If your intent is to add a well, use add_well instead."
            )
        path = self._add_image(
            row=row,
            column=column,
            image_path=image_path,
            acquisition_id=acquisition_id,
            acquisition_name=acquisition_name,
            atomic=False,
        )
        return path

    def add_well(
        self,
        row: str,
        column: int | str,
    ) -> OmeZarrWell:
        """Add a well to an ome-zarr plate."""
        _ = self._add_image(
            row=row,
            column=column,
            image_path=None,
            acquisition_id=None,
            acquisition_name=None,
            atomic=False,
        )
        return self.get_well(row=row, column=column)

    def add_column(
        self,
        column: int | str,
    ) -> "OmeZarrPlate":
        """Add a column to an ome-zarr plate."""
        meta, _ = self.meta.add_column(column)
        self.meta_handler.write_meta(meta)
        self.meta_handler._group_handler.clean_cache()
        return self

    def add_row(
        self,
        row: str,
    ) -> "OmeZarrPlate":
        """Add a row to an ome-zarr plate."""
        meta, _ = self.meta.add_row(row)
        self.meta_handler.write_meta(meta)
        self.meta_handler._group_handler.clean_cache()
        return self

    def add_acquisition(
        self,
        acquisition_id: int,
        acquisition_name: str,
    ) -> "OmeZarrPlate":
        """Add an acquisition to an ome-zarr plate.

        Be aware that this is not a parallel safe operation.

        Args:
            acquisition_id (int): The acquisition id.
            acquisition_name (str): The acquisition name.
        """
        meta = self.meta.add_acquisition(
            acquisition_id=acquisition_id, acquisition_name=acquisition_name
        )
        self.meta_handler.write_meta(meta)
        self.meta_handler._group_handler.clean_cache()
        return self

    def _remove_well(
        self,
        row: str,
        column: int | str,
        atomic: bool = False,
    ):
        """Remove a well from an ome-zarr plate."""
        if atomic:
            plate_lock = self._group_handler.lock
        else:
            plate_lock = MockLock()

        with plate_lock:
            meta = self.meta
            meta = meta.remove_well(row, column)
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

    def _remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        atomic: bool = False,
    ):
        """Remove an image from an ome-zarr plate."""
        well = self.get_well(row, column)

        if atomic:
            well_lock = well.meta_handler._group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            well_meta = well.meta
            well_meta = well_meta.remove_image(path=image_path)
            well.meta_handler.write_meta(well_meta)
            well.meta_handler._group_handler.clean_cache()
            if len(well_meta.paths()) == 0:
                self._remove_well(row, column, atomic=atomic)

    def atomic_remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
    ):
        """Parallel safe version of remove_image."""
        return self._remove_image(
            row=row,
            column=column,
            image_path=image_path,
            atomic=True,
        )

    def remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
    ):
        """Remove an image from an ome-zarr plate."""
        return self._remove_image(
            row=row,
            column=column,
            image_path=image_path,
            atomic=False,
        )

    def derive_plate(
        self,
        store: StoreOrGroup,
        plate_name: str | None = None,
        version: NgffVersions = "0.4",
        keep_acquisitions: bool = False,
        cache: bool = False,
        overwrite: bool = False,
        parallel_safe: bool = True,
    ) -> "OmeZarrPlate":
        """Derive a new OME-Zarr plate from an existing one.

        Args:
            store (StoreOrGroup): The Zarr store or group that stores the plate.
            plate_name (str | None): The name of the new plate.
            version (NgffVersion): The version of the new plate.
            keep_acquisitions (bool): Whether to keep the acquisitions in the new plate.
            cache (bool): Whether to use a cache for the zarr group metadata.
            overwrite (bool): Whether to overwrite the existing plate.
            parallel_safe (bool): Whether the group handler is parallel safe.
        """
        return derive_ome_zarr_plate(
            ome_zarr_plate=self,
            store=store,
            plate_name=plate_name,
            version=version,
            keep_acquisitions=keep_acquisitions,
            cache=cache,
            overwrite=overwrite,
            parallel_safe=parallel_safe,
        )

    @property
    def tables_container(self) -> TablesContainer:
        """Return the tables container."""
        if self._tables_container is None:
            self._tables_container = _default_table_container(self._group_handler)
            if self._tables_container is None:
                raise NgioValidationError("No tables found in the image.")
        return self._tables_container

    @property
    def list_tables(self) -> list[str]:
        """List all tables in the image."""
        return self.tables_container.list()

    def list_roi_tables(self) -> list[str]:
        """List all ROI tables in the image."""
        return self.tables_container.list_roi_tables()

    @overload
    def get_table(self, name: str) -> Table: ...

    @overload
    def get_table(self, name: str, check_type: None) -> Table: ...

    @overload
    def get_table(self, name: str, check_type: Literal["roi_table"]) -> RoiTable: ...

    @overload
    def get_table(
        self, name: str, check_type: Literal["masking_roi_table"]
    ) -> MaskingRoiTable: ...

    @overload
    def get_table(
        self, name: str, check_type: Literal["feature_table"]
    ) -> FeatureTable: ...

    @overload
    def get_table(
        self, name: str, check_type: Literal["generic_roi_table"]
    ) -> GenericRoiTable: ...

    def get_table(self, name: str, check_type: TypedTable | None = None) -> Table:
        """Get a table from the image.

        Args:
            name (str): The name of the table.
            check_type (TypedTable | None): The type of the table. If None, the
                type is not checked. If a type is provided, the table must be of that
                type.
        """
        if check_type is None:
            table = self.tables_container.get(name, strict=False)
            return table

        table = self.tables_container.get(name, strict=True)
        match check_type:
            case "roi_table":
                if not isinstance(table, RoiTable):
                    raise NgioValueError(
                        f"Table '{name}' is not a ROI table. Found type: {table.type()}"
                    )
                return table
            case "masking_roi_table":
                if not isinstance(table, MaskingRoiTable):
                    raise NgioValueError(
                        f"Table '{name}' is not a masking ROI table. "
                        f"Found type: {table.type()}"
                    )
                return table

            case "generic_roi_table":
                if not isinstance(table, GenericRoiTable):
                    raise NgioValueError(
                        f"Table '{name}' is not a generic ROI table. "
                        f"Found type: {table.type()}"
                    )
                return table

            case "feature_table":
                if not isinstance(table, FeatureTable):
                    raise NgioValueError(
                        f"Table '{name}' is not a feature table. "
                        f"Found type: {table.type()}"
                    )
                return table
            case _:
                raise NgioValueError(f"Unknown check_type: {check_type}")

    def add_table(
        self,
        name: str,
        table: Table,
        backend: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """Add a table to the image."""
        self.tables_container.add(
            name=name, table=table, backend=backend, overwrite=overwrite
        )


def open_ome_zarr_plate(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Open an OME-Zarr plate.

    Args:
        store (StoreOrGroup): The Zarr store or group that stores the plate.
        cache (bool): Whether to use a cache for the zarr group metadata.
        mode (AccessModeLiteral): The
            access mode for the image. Defaults to "r+".
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    group_handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    return OmeZarrPlate(group_handler)


def _create_empty_plate_from_meta(
    store: StoreOrGroup,
    meta: NgioPlateMeta,
    version: NgffVersions = "0.4",
    overwrite: bool = False,
) -> ZarrGroupHandler:
    """Create an empty OME-Zarr plate from metadata."""
    mode = "w" if overwrite else "w-"
    group_handler = ZarrGroupHandler(
        store=store, cache=True, mode=mode, parallel_safe=False
    )
    meta_handler = get_plate_meta_handler(group_handler, version=version)
    meta_handler.write_meta(meta)
    return group_handler


def create_empty_plate(
    store: StoreOrGroup,
    name: str,
    images: list[ImageInWellPath] | None = None,
    version: NgffVersions = "0.4",
    cache: bool = False,
    overwrite: bool = False,
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Initialize and create an empty OME-Zarr plate."""
    plate_meta = NgioPlateMeta.default_init(
        name=name,
        version=version,
    )
    group_handler = _create_empty_plate_from_meta(
        store=store,
        meta=plate_meta,
        version=version,
        overwrite=overwrite,
    )

    if images is not None:
        plate = OmeZarrPlate(group_handler)
        for image in images:
            plate.add_image(
                row=image.row,
                column=image.column,
                image_path=image.path,
                acquisition_id=image.acquisition_id,
                acquisition_name=image.acquisition_name,
            )
    return open_ome_zarr_plate(
        store=store,
        cache=cache,
        mode="r+",
        parallel_safe=parallel_safe,
    )


def derive_ome_zarr_plate(
    ome_zarr_plate: OmeZarrPlate,
    store: StoreOrGroup,
    plate_name: str | None = None,
    version: NgffVersions = "0.4",
    keep_acquisitions: bool = False,
    cache: bool = False,
    overwrite: bool = False,
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Derive a new OME-Zarr plate from an existing one.

    Args:
        ome_zarr_plate (OmeZarrPlate): The existing OME-Zarr plate.
        store (StoreOrGroup): The Zarr store or group that stores the plate.
        plate_name (str | None): The name of the new plate.
        version (NgffVersion): The version of the new plate.
        keep_acquisitions (bool): Whether to keep the acquisitions in the new plate.
        cache (bool): Whether to use a cache for the zarr group metadata.
        overwrite (bool): Whether to overwrite the existing plate.
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    if plate_name is None:
        plate_name = ome_zarr_plate.meta.plate.name

    new_meta = ome_zarr_plate.meta.derive(
        name=plate_name,
        version=version,
        keep_acquisitions=keep_acquisitions,
    )
    _ = _create_empty_plate_from_meta(
        store=store,
        meta=new_meta,
        overwrite=overwrite,
        version=version,
    )
    return open_ome_zarr_plate(
        store=store,
        cache=cache,
        mode="r+",
        parallel_safe=parallel_safe,
    )


def open_ome_zarr_well(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
    parallel_safe: bool = True,
) -> OmeZarrWell:
    """Open an OME-Zarr well.

    Args:
        store (StoreOrGroup): The Zarr store or group that stores the plate.
        cache (bool): Whether to use a cache for the zarr group metadata.
        mode (AccessModeLiteral): The access mode for the image. Defaults to "r+".
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    group_handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    return OmeZarrWell(group_handler)


def create_empty_well(
    store: StoreOrGroup,
    version: NgffVersions = "0.4",
    cache: bool = False,
    overwrite: bool = False,
    parallel_safe: bool = True,
) -> OmeZarrWell:
    """Create an empty OME-Zarr well.

    Args:
        store (StoreOrGroup): The Zarr store or group that stores the well.
        version (NgffVersion): The version of the new well.
        cache (bool): Whether to use a cache for the zarr group metadata.
        overwrite (bool): Whether to overwrite the existing well.
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    group_handler = ZarrGroupHandler(
        store=store, cache=True, mode="w" if overwrite else "w-", parallel_safe=False
    )
    meta_handler = get_well_meta_handler(group_handler, version=version)
    meta = NgioWellMeta.default_init()
    meta_handler.write_meta(meta)

    return open_ome_zarr_well(
        store=store,
        cache=cache,
        mode="r+",
        parallel_safe=parallel_safe,
    )



import pandas as pd
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame

from ngio.tables.backends._abstract_backend import AbstractTableBackend
from ngio.tables.backends._utils import (
    normalize_pandas_df,
    normalize_polars_lf,
)
from ngio.utils import NgioFileNotFoundError


class JsonTableBackend(AbstractTableBackend):
    """A class to load and write small tables in the zarr group .attrs (json)."""

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        return "experimental_json_v1"

    @staticmethod
    def implements_anndata() -> bool:
        """Whether the handler implements the anndata protocol."""
        return False

    @staticmethod
    def implements_pandas() -> bool:
        """Whether the handler implements the dataframe protocol."""
        return True

    @staticmethod
    def implements_polars() -> bool:
        """Whether the handler implements the polars protocol."""
        return True

    def _get_table_group(self):
        """Get the table group, creating it if it doesn't exist."""
        try:
            table_group = self._group_handler.get_group(path="table")
        except NgioFileNotFoundError:
            table_group = self._group_handler.group.create_group("table")
        return table_group

    def _load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        table_group = self._get_table_group()
        table_dict = dict(table_group.attrs)

        data_frame = pd.DataFrame.from_dict(table_dict)
        return data_frame

    def load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        data_frame = self._load_as_pandas_df()
        data_frame = normalize_pandas_df(
            data_frame,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=False,
        )
        return data_frame

    def _write_from_dict(self, table: dict) -> None:
        """Write the table from a dictionary to the store."""
        table_group = self._get_table_group()
        table_group.attrs.clear()
        table_group.attrs.update(table)

    def write_from_pandas(self, table: DataFrame) -> None:
        """Write the table from a pandas DataFrame."""
        table = normalize_pandas_df(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=True,
        )
        table_dict = table.to_dict(orient="list")
        self._write_from_dict(table=table_dict)

    def write_from_polars(self, table: PolarsDataFrame | LazyFrame) -> None:
        """Write the table from a polars DataFrame or LazyFrame."""
        table = normalize_polars_lf(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
        )
        if isinstance(table, LazyFrame):
            table = table.collect()

        table_dict = table.to_dict(as_series=False)
        self._write_from_dict(table=table_dict)



"""Download testing OME-Zarr datasets."""

from pathlib import Path

import pooch

from ngio.utils._errors import NgioValueError

_ome_zarr_zoo = {
    "CardiomyocyteTiny": {
        "url": "https://zenodo.org/records/13305156/files/20200812-CardiomyocyteDifferentiation14-Cycle1.zarr.zip",
        "known_hash": "md5:efc21fe8d4ea3abab76226d8c166452c",
        "fname": "20200812-CardiomyocyteDifferentiation14-Cycle1.zarr.zip",
        "processor": pooch.Unzip(extract_dir=""),
    },
    "CardiomyocyteSmallMip": {
        "url": "https://zenodo.org/records/13305316/files/20200812-CardiomyocyteDifferentiation14-Cycle1_mip.zarr.zip",
        "known_hash": "md5:3ed3ea898e0ed42d397da2e1dbe40750",
        "fname": "20200812-CardiomyocyteDifferentiation14-Cycle1_mip.zarr.zip",
        "processor": pooch.Unzip(extract_dir=""),
    },
}


def list_ome_zarr_datasets() -> list[str]:
    """List available OME-Zarr datasets."""
    return list(_ome_zarr_zoo.keys())


def download_ome_zarr_dataset(
    dataset_name: str,
    download_dir: str | Path = "data",
) -> Path:
    """Download an OME-Zarr dataset.

    To list available datasets, use `list_ome_zarr_datasets`.

    Args:
        dataset_name (str): The dataset name.
        download_dir (str): The download directory. Defaults to "data".
    """
    if dataset_name not in _ome_zarr_zoo:
        raise NgioValueError(f"Dataset {dataset_name} not found in the OME-Zarr zoo.")
    ome_zarr_url = _ome_zarr_zoo[dataset_name]
    pooch.retrieve(
        path=download_dir,
        **ome_zarr_url,
    )
    path = Path(download_dir) / ome_zarr_url["fname"]

    if isinstance(ome_zarr_url["processor"], pooch.Unzip):
        path = path.with_suffix("")
    return path



from functools import partial

import pytest
from pandas import DataFrame

from ngio.tables._validators import (
    validate_columns,
    validate_table,
    validate_unique_index,
)
from ngio.utils import NgioTableValidationError


def test_validate_columns():
    test_df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]})
    required_columns = ["a", "b"]
    optional_columns = ["c"]
    result = validate_columns(test_df, required_columns, optional_columns)
    assert result.equals(test_df)

    with pytest.raises(NgioTableValidationError):
        validate_columns(test_df, ["a", "b", "d"], optional_columns)

    with pytest.raises(NgioTableValidationError):
        validate_columns(test_df, ["a", "b"], optional_columns=["d"])


def test_validate_index():
    test_df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    result = validate_unique_index(test_df)
    assert result.equals(test_df)

    test_df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}, index=[1, 1, 3])
    with pytest.raises(NgioTableValidationError):
        validate_unique_index(test_df)


def test_validate_table():
    test_df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]})
    required_columns = ["a", "b"]
    optional_columns = ["c"]

    validators = [
        partial(
            validate_columns,
            required_columns=required_columns,
            optional_columns=optional_columns,
        ),
        validate_unique_index,
    ]

    out_df = validate_table(test_df, validators)
    assert out_df.equals(test_df)



from pathlib import Path

from ngio.ome_zarr_meta import NgioImageMeta, find_image_meta_handler
from ngio.utils import ZarrGroupHandler


def test_get_image_handler(cardiomyocyte_tiny_path: Path):
    # TODO this is a placeholder test
    # The pooch cache is giving us trouble here
    cardiomyocyte_tiny_path = cardiomyocyte_tiny_path / "B" / "03" / "0"
    group_handler = ZarrGroupHandler(cardiomyocyte_tiny_path)
    handler = find_image_meta_handler(group_handler)
    meta = handler.safe_load_meta()
    assert isinstance(meta, NgioImageMeta)
    handler.write_meta(meta)



from typing import TypeVar

import dask.array as da
import numpy as np

from ngio.ome_zarr_meta.ngio_specs._axes import (
    AxesExpand,
    AxesSqueeze,
    AxesTransformation,
    AxesTranspose,
)
from ngio.utils import NgioValueError

T = TypeVar("T")


def transform_list(
    input_list: list[T], default: T, operations: tuple[AxesTransformation, ...]
) -> list[T]:
    if isinstance(input_list, tuple):
        input_list = list(input_list)

    for operation in operations:
        if isinstance(operation, AxesTranspose):
            input_list = [input_list[i] for i in operation.axes]

        if isinstance(operation, AxesExpand):
            for ax in operation.axes:
                input_list.insert(ax, default)
        elif isinstance(operation, AxesSqueeze):
            for offset, ax in enumerate(operation.axes):
                input_list.pop(ax - offset)

    return input_list


def transform_numpy_array(
    array: np.ndarray, operations: tuple[AxesTransformation, ...]
) -> np.ndarray:
    for operation in operations:
        if isinstance(operation, AxesTranspose):
            array = np.transpose(array, operation.axes)
        elif isinstance(operation, AxesExpand):
            array = np.expand_dims(array, axis=operation.axes)
        elif isinstance(operation, AxesSqueeze):
            array = np.squeeze(array, axis=operation.axes)
        else:
            raise NgioValueError(f"Unknown operation {operation}")
    return array


def transform_dask_array(
    array: da.Array, operations: tuple[AxesTransformation, ...]
) -> da.Array:
    for operation in operations:
        if isinstance(operation, AxesTranspose):
            array = da.transpose(array, axes=operation.axes)
        elif isinstance(operation, AxesExpand):
            array = da.expand_dims(array, axis=operation.axes)
        elif isinstance(operation, AxesSqueeze):
            array = da.squeeze(array, axis=operation.axes)
        else:
            raise NgioValueError(f"Unknown operation {operation}")
    return array



import io

import pandas as pd
import polars as pl
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame
from zarr.storage import DirectoryStore, FSStore

from ngio.tables.backends._abstract_backend import AbstractTableBackend
from ngio.tables.backends._utils import normalize_pandas_df, normalize_polars_lf
from ngio.utils import NgioFileNotFoundError, NgioValueError


class CsvTableBackend(AbstractTableBackend):
    """A class to load and write small tables in CSV format."""

    csv_name = "table.csv"

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        return "experimental_csv_v1"

    @staticmethod
    def implements_anndata() -> bool:
        """Whether the handler implements the anndata protocol."""
        return False

    @staticmethod
    def implements_pandas() -> bool:
        """Whether the handler implements the dataframe protocol."""
        return True

    @staticmethod
    def implements_polars() -> bool:
        """Whether the handler implements the polars protocol."""
        return True

    def _load_from_directory_store(self, reader):
        """Load the table from a directory store."""
        url = self._group_handler.full_url
        if url is None:
            raise NgioValueError(
                f"Ngio does not support reading a CSV file from a "
                f"store of type {type(self._group_handler)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore."
            )
        csv_path = f"{url}/{self.csv_name}"
        dataframe = reader(csv_path)
        return dataframe

    def _load_from_fs_store(self, reader):
        """Load the table from an FS store."""
        bytes_table = self._group_handler.store.get(self.csv_name)
        if bytes_table is None:
            raise NgioFileNotFoundError(f"No table found at {self.csv_name}. ")
        dataframe = reader(io.BytesIO(bytes_table))
        return dataframe

    def load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        store = self._group_handler.store
        if isinstance(store, DirectoryStore):
            dataframe = self._load_from_directory_store(reader=pd.read_csv)
        elif isinstance(store, FSStore):
            dataframe = self._load_from_fs_store(reader=pd.read_csv)
        else:
            raise NgioValueError(
                f"Ngio does not support reading a CSV file from a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )

        dataframe = normalize_pandas_df(
            dataframe,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=False,
        )
        return dataframe

    def load_as_polars_lf(self) -> LazyFrame:
        """Load the table as a polars LazyFrame."""
        store = self._group_handler.store
        if isinstance(store, DirectoryStore):
            lazy_frame = self._load_from_directory_store(reader=pl.scan_csv)
        elif isinstance(store, FSStore):
            lazy_frame = self._load_from_fs_store(reader=pl.scan_csv)
        else:
            raise NgioValueError(
                f"Ngio does not support reading a CSV file from a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )
        if not isinstance(lazy_frame, LazyFrame):
            raise NgioValueError(
                "Table is not a lazy frame. Please report this issue as an ngio bug."
                f" {type(lazy_frame)}"
            )

        lazy_frame = normalize_polars_lf(
            lazy_frame,
            index_key=self.index_key,
            index_type=self.index_type,
        )
        return lazy_frame

    def _get_store_url(self) -> str:
        """Get the store URL."""
        store = self._group_handler.store
        if isinstance(store, DirectoryStore):
            full_url = self._group_handler.full_url
        else:
            raise NgioValueError(
                f"Ngio does not support writing a CSV file to a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )
        if full_url is None:
            raise NgioValueError(
                f"Ngio does not support writing a CSV file to a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )
        return full_url

    def write_from_pandas(self, table: DataFrame) -> None:
        """Write the table from a pandas DataFrame."""
        table = normalize_pandas_df(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=True,
        )
        full_url = self._get_store_url()
        csv_path = f"{full_url}/{self.csv_name}"
        table.to_csv(csv_path, index=False)

    def write_from_polars(self, table: PolarsDataFrame | LazyFrame) -> None:
        """Write the table from a polars DataFrame or LazyFrame."""
        table = normalize_polars_lf(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
        )

        if isinstance(table, LazyFrame):
            table = table.collect()

        full_url = self._get_store_url()
        csv_path = f"{full_url}/{self.csv_name}"
        table.write_csv(csv_path)



"""Region of interest (ROI) metadata.

These are the interfaces bwteen the ROI tables / masking ROI tables and
    the ImageLikeHandler.
"""

from collections.abc import Iterable

import numpy as np
from pydantic import BaseModel, ConfigDict, Field

from ngio.common._dimensions import Dimensions
from ngio.ome_zarr_meta.ngio_specs import DefaultSpaceUnit, PixelSize, SpaceUnits
from ngio.utils import NgioValueError


def _to_raster(value: float, pixel_size: float, max_shape: int) -> int:
    """Convert to raster coordinates."""
    round_value = int(np.round(value / pixel_size))
    # Ensure the value is within the image shape boundaries
    return max(0, min(round_value, max_shape))


def _to_world(value: int, pixel_size: float) -> float:
    """Convert to world coordinates."""
    return value * pixel_size


class Roi(BaseModel):
    """Region of interest (ROI) metadata."""

    name: str
    x_length: float
    y_length: float
    z_length: float = 1.0
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    unit: SpaceUnits | str | None = Field(DefaultSpaceUnit, repr=False)

    model_config = ConfigDict(extra="allow")

    def to_pixel_roi(
        self, pixel_size: PixelSize, dimensions: Dimensions
    ) -> "RoiPixels":
        """Convert to raster coordinates."""
        dim_x = dimensions.get("x")
        dim_y = dimensions.get("y")
        # Will default to 1 if z does not exist
        dim_z = dimensions.get("z", strict=False)

        return RoiPixels(
            name=self.name,
            x=_to_raster(self.x, pixel_size.x, dim_x),
            y=_to_raster(self.y, pixel_size.y, dim_y),
            z=_to_raster(self.z, pixel_size.z, dim_z),
            x_length=_to_raster(self.x_length, pixel_size.x, dim_x),
            y_length=_to_raster(self.y_length, pixel_size.y, dim_y),
            z_length=_to_raster(self.z_length, pixel_size.z, dim_z),
        )

    def zoom(self, zoom_factor: float = 1) -> "Roi":
        """Zoom the ROI by a factor.

        Args:
            zoom_factor: The zoom factor. If the zoom factor
                is less than 1 the ROI will be zoomed in.
                If the zoom factor is greater than 1 the ROI will be zoomed out.
                If the zoom factor is 1 the ROI will not be changed.
        """
        return zoom_roi(self, zoom_factor)


class RoiPixels(BaseModel):
    """Region of interest (ROI) metadata."""

    name: str
    x: int
    y: int
    z: int
    x_length: int
    y_length: int
    z_length: int
    model_config = ConfigDict(extra="allow")

    def to_roi(self, pixel_size: PixelSize) -> Roi:
        """Convert to world coordinates."""
        return Roi(
            name=self.name,
            x=_to_world(self.x, pixel_size.x),
            y=_to_world(self.y, pixel_size.y),
            z=_to_world(self.z, pixel_size.z),
            x_length=_to_world(self.x_length, pixel_size.x),
            y_length=_to_world(self.y_length, pixel_size.y),
            z_length=_to_world(self.z_length, pixel_size.z),
            unit=pixel_size.space_unit,
        )

    def to_slices(self) -> dict[str, slice]:
        """Return the slices for the ROI."""
        return {
            "x": slice(self.x, self.x + self.x_length),
            "y": slice(self.y, self.y + self.y_length),
            "z": slice(self.z, self.z + self.z_length),
        }


def zoom_roi(roi: Roi, zoom_factor: float = 1) -> Roi:
    """Zoom the ROI by a factor.

    Args:
        roi: The ROI to zoom.
        zoom_factor: The zoom factor. If the zoom factor
            is less than 1 the ROI will be zoomed in.
            If the zoom factor is greater than 1 the ROI will be zoomed out.
            If the zoom factor is 1 the ROI will not be changed.
    """
    if zoom_factor <= 0:
        raise ValueError("Zoom factor must be greater than 0.")

    # the zoom factor needs to be rescaled
    # from the range [-1, inf) to [0, inf)
    zoom_factor -= 1
    diff_x = roi.x_length * zoom_factor
    diff_y = roi.y_length * zoom_factor

    new_x = max(roi.x - diff_x / 2, 0)
    new_y = max(roi.y - diff_y / 2, 0)

    new_roi = Roi(
        name=roi.name,
        x=new_x,
        y=new_y,
        z=roi.z,
        x_length=roi.x_length + diff_x,
        y_length=roi.y_length + diff_y,
        z_length=roi.z_length,
        unit=roi.unit,
    )

    return new_roi


def roi_to_slice_kwargs(
    roi: Roi,
    pixel_size: PixelSize,
    dimensions: Dimensions,
    **slice_kwargs: slice | int | Iterable[int],
) -> dict[str, slice | int | Iterable[int]]:
    """Convert a WorldCooROI to slice_kwargs."""
    raster_roi = roi.to_pixel_roi(
        pixel_size=pixel_size, dimensions=dimensions
    ).to_slices()

    if not dimensions.has_axis(axis_name="z"):
        raster_roi.pop("z")

    for key in slice_kwargs.keys():
        if key in raster_roi:
            raise NgioValueError(
                f"Key {key} is already in the slice_kwargs. "
                "Ambiguous which one to use: "
                f"{key}={slice_kwargs[key]} or roi_{key}={raster_roi[key]}"
            )
    return {**raster_roi, **slice_kwargs}
